/*
 * cdc-acm.c
 *
 * Copyright (c) 1999 Armin Fuerst	<fuerst@in.tum.de>
 * Copyright (c) 1999 Pavel Machek	<pavel@suse.cz>
 * Copyright (c) 1999 Johannes Erdfelt	<johannes@erdfelt.com>
 * Copyright (c) 2000 Vojtech Pavlik	<vojtech@suse.cz>
 * Copyright (c) 2004 Oliver Neukum	<oliver@neukum.name>
 * Copyright (c) 2005 David Kubicek	<dave@awk.cz>
 * Copyright (c) 2007 Konji Suzuki	<suz@nmj.sakura.ne.jp>
 *
 * USB Abstract Control Model driver for USB modems and ISDN adapters
 *
 * Sponsored by SuSE
 *
 * ChangeLog:
 *	v0.9  - thorough cleaning, URBification, almost a rewrite
 *	v0.10 - some more cleanups
 *	v0.11 - fixed flow control, read error doesn't stop reads
 *	v0.12 - added TIOCM ioctls, added break handling, made struct acm kmalloced
 *	v0.13 - added termios, added hangup
 *	v0.14 - sized down struct acm
 *	v0.15 - fixed flow control again - characters could be lost
 *	v0.16 - added code for modems with swapped data and control interfaces
 *	v0.17 - added new style probing
 *	v0.18 - fixed new style probing for devices with more configurations
 *	v0.19 - fixed CLOCAL handling (thanks to Richard Shih-Ping Chan)
 *	v0.20 - switched to probing on interface (rather than device) class
 *	v0.21 - revert to probing on device for devices with multiple configs
 *	v0.22 - probe only the control interface. if usbcore doesn't choose the
 *		config we want, sysadmin changes bConfigurationValue in sysfs.
 *	v0.23 - use softirq for rx processing, as needed by tty layer
 *	v0.24 - change probe method to evaluate CDC union descriptor
 *	v0.25 - downstream tasks paralelized to maximize throughput
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#undef DEBUG

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/module.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/smp_lock.h>
#endif
#include <linux/mutex.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
#include <linux/usb/cdc.h>
#else
#include <linux/usb_cdc.h>
#endif
#include <asm/byteorder.h>
#include <asm/unaligned.h>
#include <linux/list.h>

#include "cdc-acm.h"

#define dbg2(format, arg...) printk(KERN_INFO ": " format "\n" , ## arg)
#if 0
#define dbgX(format, arg...) printk(KERN_INFO ": " format "\n" , ## arg)
#else
#define dbgX(format, arg...)
#endif
#define dbgY(format, arg...) printk(KERN_INFO ": " format "\n" , ## arg)

/*
 * Version Information
 */
#define DRIVER_VERSION "v0.25t"
#define DRIVER_AUTHOR "Armin Fuerst, Pavel Machek, Johannes Erdfelt, Vojtech Pavlik, David Kubicek"
#define DRIVER_DESC "USB Abstract Control Model driver for USB modems and ISDN adapters"

static struct usb_driver acm_driver;
static struct tty_driver *acm_tty_driver;
static struct acm *acm_table[ACM_TTY_MINORS];

static DEFINE_MUTEX(open_mutex);

#define ACM_READY(acm)	(acm && acm->dev && acm->used)

static struct urb *acm_get_write_urb(struct acm *acm) {
	unsigned long flags;
	int i;
	struct urb *urb = NULL;

	spin_lock_irqsave (&acm->write_lock, flags);
	for (i=0; i< ACM_NWU; i++) {
		if (!(acm->write_urb_used & (1U<<i))) {
			acm->write_urb_used |= (1U<<i);
			urb = &acm->write_urb[i];
			break;
		}
	}
	spin_unlock_irqrestore (&acm->write_lock, flags);
	if (urb && (urb->status == -EINPROGRESS)) {
		dbg("acm_get_write_urb: urb is INPROGRESS");
	}
	return urb;
}

static void acm_put_write_urb(struct acm *acm, struct urb *urb) {
	unsigned long flags;
	int i;

	i = urb - &acm->write_urb[0];
	if ((i < 0) || (i >= ACM_NWU)) {
		dbg("acm_put_write_urb: address error");
		return;
	}
	if (urb->status == -EINPROGRESS) {
		dbg("acm_put_write_urb: urb is INPROGRESS");
	}
	spin_lock_irqsave (&acm->write_lock, flags);
	acm->write_urb_used &= ~(1U<<i);
	spin_unlock_irqrestore (&acm->write_lock, flags);
}

static struct urb *acm_get_read_urb(struct acm *acm) {
	unsigned long flags;
	int next;
	struct urb *urb = NULL;

	spin_lock_irqsave (&acm->read_lock, flags);
	next = (acm->read_urb_get_idx + 1) % ACM_NRU;
	if (next == acm->read_urb_put_idx) {
		spin_unlock_irqrestore (&acm->read_lock, flags);
		return NULL;
	}

	urb = &acm->read_urb[acm->read_urb_get_idx];
	acm->read_urb_get_idx = next;
	spin_unlock_irqrestore (&acm->read_lock, flags);

	return urb;
}

static void acm_put_read_urb(struct acm *acm, struct urb *urb) {
	unsigned long flags;
	int idx;

	idx = urb - &acm->read_urb[0];
	if ((idx < 0) || (idx >= ACM_NRU)) {
		dbg("acm_put_read_urb: address error");
		return;
	}
	spin_lock_irqsave (&acm->read_lock, flags);
	if (idx != acm->read_urb_put_idx) {
		dbg("acm_put_read_urb: sequence error");
	}
	acm->read_urb_put_idx = (acm->read_urb_put_idx + 1) % ACM_NRU;
	spin_unlock_irqrestore (&acm->read_lock, flags);
}

static struct urb *acm_first_read_urb(struct acm *acm) {
	unsigned long flags;
	int nr_used;
	struct urb *urb;

	spin_lock_irqsave (&acm->read_lock, flags);
	urb =  &acm->read_urb[acm->read_urb_put_idx];
	nr_used = (acm->read_urb_get_idx - acm->read_urb_put_idx + ACM_NRU)
			% ACM_NRU;
	if (!nr_used || 
	     (nr_used - atomic_read(&acm->read_urb_outstanding) <= 0) )
		urb = NULL;
	spin_unlock_irqrestore (&acm->read_lock, flags);
	return urb;
}

static void acm_kill_read_urb(struct acm *acm) {
	struct urb *urb;
	while (atomic_read(&acm->in_read_proc)) {
		schedule_timeout(1);
	}
	while (acm->read_urb_put_idx != acm->read_urb_get_idx) {
		urb =  &acm->read_urb[acm->read_urb_put_idx];
		acm_put_read_urb(acm,urb);
		usb_kill_urb(urb);
	}
	atomic_set(&acm->read_urb_outstanding, 0);
}

/*
 * Functions for ACM control messages.
 */

static int acm_ctrl_msg(struct acm *acm, int request, int value, void *buf, int len)
{
	int retval = usb_control_msg(acm->dev, usb_sndctrlpipe(acm->dev, 0),
		request, USB_RT_ACM, value,
		acm->control->altsetting[0].desc.bInterfaceNumber,
		buf, len, 5000);
	dbg("acm_control_msg: rq: 0x%02x val: %#x len: %#x result: %d", request, value, len, retval);
	return retval < 0 ? retval : 0;
}

/* devices aren't required to support these requests.
 * the cdc acm descriptor tells whether they do...
 */
#define acm_set_control(acm, control) \
	acm_ctrl_msg(acm, USB_CDC_REQ_SET_CONTROL_LINE_STATE, control, NULL, 0)
#define acm_set_line(acm, line) \
	acm_ctrl_msg(acm, USB_CDC_REQ_SET_LINE_CODING, 0, line, sizeof *(line))
#define acm_send_break(acm, ms) \
	acm_ctrl_msg(acm, USB_CDC_REQ_SEND_BREAK, ms, NULL, 0)

/*
 * attributes exported through sysfs
 */
static ssize_t show_caps
(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf = to_usb_interface(dev);
	struct acm *acm = usb_get_intfdata(intf);

	return sprintf(buf, "%d", acm->ctrl_caps);
}
static DEVICE_ATTR(bmCapabilities, S_IRUGO, show_caps, NULL);

static ssize_t show_country_codes
(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf = to_usb_interface(dev);
	struct acm *acm = usb_get_intfdata(intf);

	memcpy(buf, acm->country_codes, acm->country_code_size);
	return acm->country_code_size;
}

static DEVICE_ATTR(wCountryCodes, S_IRUGO, show_country_codes, NULL);

static ssize_t show_country_rel_date
(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf = to_usb_interface(dev);
	struct acm *acm = usb_get_intfdata(intf);

	return sprintf(buf, "%d", acm->country_rel_date);
}

static DEVICE_ATTR(iCountryCodeRelDate, S_IRUGO, show_country_rel_date, NULL);
/*
 * Interrupt handlers for various ACM device responses
 */

/* control interface reports status changes with "interrupt" transfers */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void acm_ctrl_irq(struct urb *urb)
#else
static void acm_ctrl_irq(struct urb *urb, struct pt_regs *regs)
#endif
{
	struct acm *acm = urb->context;
	struct usb_cdc_notification *dr = urb->transfer_buffer;
	unsigned char *data;
	int newctrl;
	int retval;
	int status = urb->status;

	switch (status) {
	case 0:
		/* success */
		break;
	case -ECONNRESET:
	case -ENOENT:
	case -ESHUTDOWN:
		/* this urb is terminated, clean up */
		dbg("%s - urb shutting down with status: %d", __FUNCTION__, status);
		return;
	default:
		dbg("%s - nonzero urb status received: %d", __FUNCTION__, status);
		goto exit;
	}

	if (!ACM_READY(acm))
		goto exit;

	data = (unsigned char *)(dr + 1);
	switch (dr->bNotificationType) {

		case USB_CDC_NOTIFY_NETWORK_CONNECTION:

			dbg("%s network", dr->wValue ? "connected to" : "disconnected from");
			break;

		case USB_CDC_NOTIFY_SERIAL_STATE:

			newctrl = le16_to_cpu(get_unaligned((__le16 *) data));

			if (acm->tty && !acm->clocal && (acm->ctrlin & ~newctrl & ACM_CTRL_DCD)) {
				dbg("calling hangup");
				tty_hangup(acm->tty);
			}

			acm->ctrlin = newctrl;

			dbg("input control lines: dcd%c dsr%c break%c ring%c framing%c parity%c overrun%c",
				acm->ctrlin & ACM_CTRL_DCD ? '+' : '-',	acm->ctrlin & ACM_CTRL_DSR ? '+' : '-',
				acm->ctrlin & ACM_CTRL_BRK ? '+' : '-',	acm->ctrlin & ACM_CTRL_RI  ? '+' : '-',
				acm->ctrlin & ACM_CTRL_FRAMING ? '+' : '-',	acm->ctrlin & ACM_CTRL_PARITY ? '+' : '-',
				acm->ctrlin & ACM_CTRL_OVERRUN ? '+' : '-');

			break;

		default:
			dbg("unknown notification %d received: index %d len %d data0 %d data1 %d",
				dr->bNotificationType, dr->wIndex,
				dr->wLength, data[0], data[1]);
			break;
	}
exit:
	retval = usb_submit_urb (urb, GFP_ATOMIC);
	if (retval)
		err ("%s - usb_submit_urb failed with result %d",
		     __FUNCTION__, retval);
}

static void acm_try_to_submit_read_urb(struct acm *acm) {
	struct urb *urb;
	if ((atomic_read(&acm->read_urb_outstanding) < ACM_READ_OUTSTANDING )
		&& (urb = acm_get_read_urb(acm))) {
		urb->dev = acm->dev;
		urb->actual_length = 0;
		if (usb_submit_urb(urb, GFP_ATOMIC)) {
			dbg2("failed resubmitting read urb");
		} else {
			atomic_inc(&acm->read_urb_outstanding);
		}
dbgX("RX urb#%02d s ", urb - &acm->read_urb[0]);
	}
}

static void acm_rx_tasklet(unsigned long _acm)
{
	struct acm *acm = (void *)_acm;
	struct tty_struct *tty = acm->tty;
	struct urb *urb;
	unsigned char *data;

	if (!ACM_READY(acm)) return;

	if (atomic_read(&acm->in_read_proc)) {

		dbg2("acm_read_bulk: nested called: ");
		return;
	}

	urb = acm_first_read_urb(acm);
	if (!urb || (urb->status == -EINPROGRESS))
		return;

dbgX("RX urb#%02d e (%d bytes)", urb - &acm->read_urb[0],urb->actual_length);
	atomic_inc(&acm->in_read_proc);
next_urb:
	if (urb->status) {
		dbg("nonzero read bulk status received: %d", urb->status);
		urb->status = 0;
		urb->actual_length = 0;
		goto discarded;
	}

	data = urb->transfer_buffer;
	if (!urb->status && urb->actual_length && !acm->throttle)  {
		tty_buffer_request_room(tty, urb->actual_length);
		tty_insert_flip_string(tty, data, urb->actual_length);
		tty_flip_buffer_push(tty);
	}
discarded:
	acm_put_read_urb(acm,urb);
	if (acm->throttle) {
		atomic_dec(&acm->in_read_proc);
		return;
	}
	acm_try_to_submit_read_urb(acm);

	urb = acm_first_read_urb(acm);
	if (ACM_READY(acm) &&
	    urb && (urb->status != -EINPROGRESS))
		goto next_urb;
	atomic_dec(&acm->in_read_proc);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void acm_read_done(struct urb *urb)
#else
static void acm_read_done(struct urb *urb, struct pt_regs *regs)
#endif
{
	struct acm *acm = urb->context;

	if (!ACM_READY(acm)) return;

	if (urb->status)
		dbg2("nonzero read bulk status received: %d", urb->status);

dbgX("RX urb#%02d ie ", urb - &acm->read_urb[0]);
	atomic_dec(&acm->read_urb_outstanding);
	acm_try_to_submit_read_urb(acm);
	tasklet_schedule(&acm->urb_task);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void acm_write_done(struct urb *urb)
#else
static void acm_write_done(struct urb *urb, struct pt_regs *regs)
#endif
{
	struct acm *acm = (struct acm *)urb->context;

	if (!ACM_READY(acm)) {
		if (acm && acm->dev) { // closing
			acm_put_write_urb(acm, urb);
			atomic_dec(&acm->write_urb_outstanding);
		}
		return;
	}

	if (urb->status)
		dbg2("nonzero write bulk status received: %d", urb->status);

dbgX("TX urb#%02d e ", urb - &acm->write_urb[0]);
	acm_put_write_urb(acm, urb);
	atomic_dec(&acm->write_urb_outstanding);

	schedule_work(&acm->work);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void acm_softint(struct work_struct *work)
#else
static void acm_softint(void *private)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	struct acm *acm = container_of(work, struct acm, work);
#else
	struct acm *acm = private;
#endif
	dbg("Entering acm_softint.");
	
	if (!ACM_READY(acm))
		return;
	tty_wakeup(acm->tty);
}

/*
 * TTY handlers
 */

static int acm_tty_open(struct tty_struct *tty, struct file *filp)
{
	struct acm *acm;
	int rv = -EINVAL;

	dbg("Entering acm_tty_open.");

	mutex_lock(&open_mutex);

	acm = acm_table[tty->index];
	if (!acm || !acm->dev)
		goto err_out;
	else
		rv = 0;

	tty->driver_data = acm;
	acm->tty = tty;

	/* force low_latency on so that our tty_push actually forces the data through,
	   otherwise it is scheduled, and with high data rates data can get lost. */
	tty->low_latency = 1;

	if (acm->used++) {
		goto done;
        }

	acm->ctrlurb.dev = acm->dev;
	if (usb_submit_urb(&acm->ctrlurb, GFP_KERNEL)) {
		dbg("usb_submit_urb(ctrl irq) failed");
		goto bail_out;
	}

	if (0 > acm_set_control(acm, acm->ctrlout = ACM_CTRL_DTR | ACM_CTRL_RTS) &&
	    (acm->ctrl_caps & USB_CDC_CAP_LINE))
		goto full_bailout;

	acm->throttle = 0;

	acm_try_to_submit_read_urb(acm);
done:
err_out:
	mutex_unlock(&open_mutex);
	return rv;

full_bailout:
	usb_kill_urb(&acm->ctrlurb);
bail_out:
	acm->used--;
	mutex_unlock(&open_mutex);
	return -EIO;
}

static void acm_tty_unregister(struct acm *acm)
{
	tty_unregister_device(acm_tty_driver, acm->minor);
	usb_put_intf(acm->control);
	acm_table[acm->minor] = NULL;
	kfree(acm->country_codes);
	kfree(acm);
}

static void acm_tty_close(struct tty_struct *tty, struct file *filp)
{
	struct acm *acm = tty->driver_data;

	if (!acm || !acm->used)
		return;

	mutex_lock(&open_mutex);
	if (!--acm->used) {
		if (acm->dev) {
			while (atomic_read(&acm->write_urb_outstanding)) {
				schedule_timeout(1);
			}
			acm_set_control(acm, acm->ctrlout = 0);
			usb_kill_urb(&acm->ctrlurb);
			acm_kill_read_urb(acm);
		} else
			acm_tty_unregister(acm);
	}
	mutex_unlock(&open_mutex);
}

static int acm_tty_write(struct tty_struct *tty, const unsigned char *buf, int count)
{
	struct acm *acm = tty->driver_data;
	struct urb *urb;
	int bytes_sent = 0;
	int cur_count;

	if (!count) return 0;

	while (count > 0) {
		if (!ACM_READY(acm)) return -EINVAL;

		if (atomic_read(&acm->write_urb_outstanding) >= ACM_WRITE_OUTSTANDING)
			break;
		urb = acm_get_write_urb(acm);
		if (!urb) break;
		cur_count = (count > acm->writesize) ? acm->writesize : count;
		memcpy(urb->transfer_buffer, buf, cur_count);
		urb->transfer_buffer_length = cur_count;
		urb->dev = acm->dev;

dbgX("TX urb#%02d s (%d bytes)", urb - &acm->write_urb[0],urb->transfer_buffer_length);
		if (usb_submit_urb(urb, GFP_ATOMIC)) {
			dbg2("usb_submit_urb(write bulk) failed");
			acm_put_write_urb(acm, urb);
			break;
		}
		buf += cur_count;
		bytes_sent += cur_count;
		count -= cur_count;
		atomic_inc(&acm->write_urb_outstanding);
	}
	return bytes_sent;
}

static int acm_tty_write_room(struct tty_struct *tty)
{
	struct acm *acm = tty->driver_data;
	int i,room = 0;
	unsigned long flags;

	if (!ACM_READY(acm)) return -EINVAL;
	spin_lock_irqsave (&acm->write_lock, flags);
	for (i=0; i< ACM_NWU; i++) {
		if (!(acm->write_urb_used & (1U<<i)))
			room += acm->writesize;
	}
	spin_unlock_irqrestore (&acm->write_lock, flags);
	return room;
}

static int acm_tty_chars_in_buffer(struct tty_struct *tty)
{
	struct acm *acm = tty->driver_data;
	int i,chars = 0;
	unsigned long flags;

	if (!ACM_READY(acm)) return -EINVAL;
	spin_lock_irqsave (&acm->write_lock, flags);
	for (i=0; i< ACM_NWU; i++) {
		if ((acm->write_urb_used & (1U<<i)) &&
		     acm->write_urb[i].status == -EINPROGRESS)
			chars += acm->write_urb[i].transfer_buffer_length;
	}
	spin_unlock_irqrestore (&acm->write_lock, flags);
	return chars;
}

static void acm_tty_throttle(struct tty_struct *tty)
{
	struct acm *acm = tty->driver_data;
	if (!ACM_READY(acm))
		return;
	spin_lock_bh(&acm->throttle_lock);
	acm->throttle = 1;
	spin_unlock_bh(&acm->throttle_lock);
}

static void acm_tty_unthrottle(struct tty_struct *tty)
{
	struct acm *acm = tty->driver_data;
	if (!ACM_READY(acm))
		return;
	spin_lock_bh(&acm->throttle_lock);
	acm->throttle = 0;
	spin_unlock_bh(&acm->throttle_lock);
	acm_try_to_submit_read_urb(acm);
	tasklet_schedule(&acm->urb_task);
}

static void acm_tty_break_ctl(struct tty_struct *tty, int state)
{
	struct acm *acm = tty->driver_data;
	if (!ACM_READY(acm))
		return;
	if (acm_send_break(acm, state ? 0xffff : 0))
		dbg("send break failed");
}

static int acm_tty_tiocmget(struct tty_struct *tty, struct file *file)
{
	struct acm *acm = tty->driver_data;

	if (!ACM_READY(acm))
		return -EINVAL;

	return (acm->ctrlout & ACM_CTRL_DTR ? TIOCM_DTR : 0) |
	       (acm->ctrlout & ACM_CTRL_RTS ? TIOCM_RTS : 0) |
	       (acm->ctrlin  & ACM_CTRL_DSR ? TIOCM_DSR : 0) |
	       (acm->ctrlin  & ACM_CTRL_RI  ? TIOCM_RI  : 0) |
	       (acm->ctrlin  & ACM_CTRL_DCD ? TIOCM_CD  : 0) |
	       TIOCM_CTS;
}

static int acm_tty_tiocmset(struct tty_struct *tty, struct file *file,
			    unsigned int set, unsigned int clear)
{
	struct acm *acm = tty->driver_data;
	unsigned int newctrl;

	if (!ACM_READY(acm))
		return -EINVAL;

	newctrl = acm->ctrlout;
	set = (set & TIOCM_DTR ? ACM_CTRL_DTR : 0) | (set & TIOCM_RTS ? ACM_CTRL_RTS : 0);
	clear = (clear & TIOCM_DTR ? ACM_CTRL_DTR : 0) | (clear & TIOCM_RTS ? ACM_CTRL_RTS : 0);

	newctrl = (newctrl & ~clear) | set;

	if (acm->ctrlout == newctrl)
		return 0;
	return acm_set_control(acm, acm->ctrlout = newctrl);
}

static int acm_tty_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct acm *acm = tty->driver_data;

	if (!ACM_READY(acm))
		return -EINVAL;

	return -ENOIOCTLCMD;
}

static const __u32 acm_tty_speed[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600,
	1200, 1800, 2400, 4800, 9600, 19200, 38400,
	57600, 115200, 230400, 460800, 500000, 576000,
	921600, 1000000, 1152000, 1500000, 2000000,
	2500000, 3000000, 3500000, 4000000
};

static const __u8 acm_tty_size[] = {
	5, 6, 7, 8
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void acm_tty_set_termios(struct tty_struct *tty, struct ktermios *termios_old)
#else
static void acm_tty_set_termios(struct tty_struct *tty, struct termios *termios_old)
#endif
{
	struct acm *acm = tty->driver_data;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	struct ktermios *termios = tty->termios;
#else
	struct termios *termios = tty->termios;
#endif
	struct usb_cdc_line_coding newline;
	int newctrl = acm->ctrlout;

	if (!ACM_READY(acm))
		return;

	newline.dwDTERate = cpu_to_le32p(acm_tty_speed +
		(termios->c_cflag & CBAUD & ~CBAUDEX) + (termios->c_cflag & CBAUDEX ? 15 : 0));
	newline.bCharFormat = termios->c_cflag & CSTOPB ? 2 : 0;
	newline.bParityType = termios->c_cflag & PARENB ?
		(termios->c_cflag & PARODD ? 1 : 2) + (termios->c_cflag & CMSPAR ? 2 : 0) : 0;
	newline.bDataBits = acm_tty_size[(termios->c_cflag & CSIZE) >> 4];

	acm->clocal = ((termios->c_cflag & CLOCAL) != 0);

	if (!newline.dwDTERate) {
		newline.dwDTERate = acm->line.dwDTERate;
		newctrl &= ~ACM_CTRL_DTR;
	} else  newctrl |=  ACM_CTRL_DTR;

	if (newctrl != acm->ctrlout)
		acm_set_control(acm, acm->ctrlout = newctrl);

	if (memcmp(&acm->line, &newline, sizeof newline)) {
		memcpy(&acm->line, &newline, sizeof newline);
		dbg("set line: %d %d %d %d", le32_to_cpu(newline.dwDTERate),
			newline.bCharFormat, newline.bParityType,
			newline.bDataBits);
		acm_set_line(acm, &acm->line);
	}
}

/*
 * USB probe and disconnect routines.
 */

static int acm_probe (struct usb_interface *intf,
		      const struct usb_device_id *id)
{
	struct usb_cdc_union_desc *union_header = NULL;
	struct usb_cdc_country_functional_desc *cfd = NULL;
	char *buffer = intf->altsetting->extra;
	int buflen = intf->altsetting->extralen;
	struct usb_interface *control_interface;
	struct usb_interface *data_interface;
	struct usb_endpoint_descriptor *epctrl;
	struct usb_endpoint_descriptor *epread;
	struct usb_endpoint_descriptor *epwrite;
	struct usb_device *usb_dev = interface_to_usbdev(intf);
	struct acm *acm;
	int minor;
	int ctrlsize,readsize,writesize;
	u8 *buf;
	u8 ac_management_function = 0;
	u8 call_management_function = 0;
	int call_interface_num = -1;
	int data_interface_num;
	unsigned long quirks;
	int i;
	struct urb *urb;

	/* normal quirks */
	quirks = (unsigned long)id->driver_info;

	/* handle quirks deadly to normal probing*/
	if (quirks == NO_UNION_NORMAL) {
		data_interface = usb_ifnum_to_if(usb_dev, 1);
		control_interface = usb_ifnum_to_if(usb_dev, 0);
		goto skip_normal_probe;
	}
	
	/* normal probing*/
	if (!buffer) {
		err("Wierd descriptor references\n");
		return -EINVAL;
	}

	if (!buflen) {
		if (intf->cur_altsetting->endpoint->extralen && intf->cur_altsetting->endpoint->extra) {
			dev_dbg(&intf->dev,"Seeking extra descriptors on endpoint");
			buflen = intf->cur_altsetting->endpoint->extralen;
			buffer = intf->cur_altsetting->endpoint->extra;
		} else {
			err("Zero length descriptor references\n");
			return -EINVAL;
		}
	}

	while (buflen > 0) {
		if (buffer [1] != USB_DT_CS_INTERFACE) {
			err("skipping garbage\n");
			goto next_desc;
		}

		switch (buffer [2]) {
			case USB_CDC_UNION_TYPE: /* we've found it */
				if (union_header) {
					err("More than one union descriptor, skipping ...");
					goto next_desc;
				}
				union_header = (struct usb_cdc_union_desc *)
							buffer;
				break;
			case USB_CDC_COUNTRY_TYPE: /* export through sysfs*/
				cfd = (struct usb_cdc_country_functional_desc *)buffer;
				break;
			case USB_CDC_HEADER_TYPE: /* maybe check version */ 
				break; /* for now we ignore it */ 
			case USB_CDC_ACM_TYPE:
				ac_management_function = buffer[3];
				break;
			case USB_CDC_CALL_MANAGEMENT_TYPE:
				call_management_function = buffer[3];
				call_interface_num = buffer[4];
				if ((call_management_function & 3) != 3)
					err("This device cannot do calls on its own. It is no modem.");
				break;
				
			default:
				err("Ignoring extra header, type %d, length %d", buffer[2], buffer[0]);
				break;
			}
next_desc:
		buflen -= buffer[0];
		buffer += buffer[0];
	}

	if (!union_header) {
		if (call_interface_num > 0) {
			dev_dbg(&intf->dev,"No union descriptor, using call management descriptor");
			data_interface = usb_ifnum_to_if(usb_dev, (data_interface_num = call_interface_num));
			control_interface = intf;
		} else {
			dev_dbg(&intf->dev,"No union descriptor, giving up");
			return -ENODEV;
		}
	} else {
		control_interface = usb_ifnum_to_if(usb_dev, union_header->bMasterInterface0);
		data_interface = usb_ifnum_to_if(usb_dev, (data_interface_num = union_header->bSlaveInterface0));
		if (!control_interface || !data_interface) {
			dev_dbg(&intf->dev,"no interfaces");
			return -ENODEV;
		}
	}
	
	if (data_interface_num != call_interface_num)
		dev_dbg(&intf->dev,"Seperate call control interface. That is not fully supported.");

skip_normal_probe:

	/*workaround for switched interfaces */
	if (data_interface->cur_altsetting->desc.bInterfaceClass != CDC_DATA_INTERFACE_TYPE) {
		if (control_interface->cur_altsetting->desc.bInterfaceClass == CDC_DATA_INTERFACE_TYPE) {
			struct usb_interface *t;
			dev_dbg(&intf->dev,"Your device has switched interfaces.");

			t = control_interface;
			control_interface = data_interface;
			data_interface = t;
		} else {
			return -EINVAL;
		}
	}

	/* Accept probe requests only for the control interface */
	if (intf != control_interface)
		return -ENODEV;
	
	if (usb_interface_claimed(data_interface)) { /* valid in this context */
		dev_dbg(&intf->dev,"The data interface isn't available");
		return -EBUSY;
	}


	if (data_interface->cur_altsetting->desc.bNumEndpoints < 2)
		return -EINVAL;

	epctrl = &control_interface->cur_altsetting->endpoint[0].desc;
	epread = &data_interface->cur_altsetting->endpoint[0].desc;
	epwrite = &data_interface->cur_altsetting->endpoint[1].desc;


	/* workaround for switched endpoints */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	if (!usb_endpoint_dir_in(epread)) {
#else
	if ((epread->bEndpointAddress & USB_DIR_IN) != USB_DIR_IN) {
#endif
		/* descriptors are swapped */
		struct usb_endpoint_descriptor *t;
		dev_dbg(&intf->dev,"The data interface has switched endpoints");
		
		t = epread;
		epread = epwrite;
		epwrite = t;
	}
	dbg("interfaces are valid");
	for (minor = 0; minor < ACM_TTY_MINORS && acm_table[minor]; minor++);

	if (minor == ACM_TTY_MINORS) {
		err("no more free acm devices");
		return -ENODEV;
	}

	if (!(acm = kzalloc(sizeof(struct acm), GFP_KERNEL))) {
		dev_dbg(&intf->dev, "out of memory (acm kzalloc)");
		goto alloc_fail;
	}

	ctrlsize = le16_to_cpu(epctrl->wMaxPacketSize);
	readsize = le16_to_cpu(epread->wMaxPacketSize);
	writesize = le16_to_cpu(epwrite->wMaxPacketSize);
	acm->control = control_interface;
	acm->data = data_interface;
	acm->minor = minor;
	acm->dev = usb_dev;
	acm->ctrl_caps = ac_management_function;
	acm->ctrlsize = ctrlsize;
	acm->readsize = readsize;
	acm->writesize = writesize;
	acm->urb_task.func = acm_rx_tasklet;
	acm->urb_task.data = (unsigned long) acm;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	INIT_WORK(&acm->work, acm_softint, acm);
#else
	INIT_WORK(&acm->work, acm_softint);
#endif
	spin_lock_init(&acm->throttle_lock);
	spin_lock_init(&acm->write_lock);
	spin_lock_init(&acm->read_lock);
	atomic_set(&acm->read_urb_outstanding, 0);
	atomic_set(&acm->in_read_proc, 0);

	usb_init_urb(&acm->ctrlurb);
	buf = usb_buffer_alloc(usb_dev, ctrlsize, GFP_KERNEL, &acm->ctrl_dma);
	if (!buf) {
		dev_dbg(&intf->dev, "out of memory (ctrl buffer alloc)");
		goto alloc_fail2;
	}
	acm->ctrl_buffer = buf;

	buf = usb_buffer_alloc(usb_dev, readsize * ACM_NRU, GFP_KERNEL, &acm->read_dma);
	if (!buf) {
		dev_dbg(&intf->dev, "out of memory (read buffer alloc)");
		goto alloc_fail3;
	}
	acm->read_buffer = buf;

	buf = usb_buffer_alloc(usb_dev, writesize * ACM_NWU, GFP_KERNEL, &acm->write_dma);
	if (!buf) {
		dev_dbg(&intf->dev, "out of memory (write buffer alloc)");
		goto alloc_fail4;
	}
	acm->write_buffer = buf;

	usb_set_intfdata (intf, acm);

	i = device_create_file(&intf->dev, &dev_attr_bmCapabilities);
	if (i < 0)
		goto alloc_fail5;

	if (cfd) { /* export the country data */
		acm->country_codes = kmalloc(cfd->bLength - 4, GFP_KERNEL);
		if (!acm->country_codes)
			goto skip_countries;
		acm->country_code_size = cfd->bLength - 4;
		memcpy(acm->country_codes, (u8 *)&cfd->wCountyCode0, cfd->bLength - 4);
		acm->country_rel_date = cfd->iCountryCodeRelDate;

		i = device_create_file(&intf->dev, &dev_attr_wCountryCodes);
		if (i < 0) {
			kfree(acm->country_codes);
			goto skip_countries;
		}

		i = device_create_file(&intf->dev, &dev_attr_iCountryCodeRelDate);
		if (i < 0) {
			kfree(acm->country_codes);
			goto skip_countries;
		}
	}

skip_countries:
	usb_fill_int_urb(&acm->ctrlurb, usb_dev, usb_rcvintpipe(usb_dev, epctrl->bEndpointAddress),
			 acm->ctrl_buffer, ctrlsize, acm_ctrl_irq, acm, epctrl->bInterval);
	//acm->ctrlurb.transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
	acm->ctrlurb.transfer_dma = acm->ctrl_dma;

	for (i = 0; i < ACM_NRU; i++) {
		urb = &acm->read_urb[i];
		usb_init_urb(urb);
		usb_fill_bulk_urb(urb, usb_dev,
			usb_rcvbulkpipe(usb_dev, epread->bEndpointAddress),
			acm->read_buffer + readsize * i,
			readsize, acm_read_done, acm);
		//urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
		urb->transfer_dma = acm->read_dma;
	}
	for (i = 0; i < ACM_NWU; i++) {
		urb = &acm->write_urb[i];
		usb_init_urb(urb);
		usb_fill_bulk_urb(urb, usb_dev,
			usb_sndbulkpipe(usb_dev, epwrite->bEndpointAddress),
			acm->write_buffer + writesize * i,
			writesize, acm_write_done, acm);
		//urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
		urb->transfer_dma = acm->write_dma;
	}

	dev_info(&intf->dev, "ttyACM%d: USB ACM device\n", minor);

	acm_set_control(acm, acm->ctrlout);

	acm->line.dwDTERate = cpu_to_le32(9600);
	acm->line.bDataBits = 8;
	acm_set_line(acm, &acm->line);

	usb_driver_claim_interface(&acm_driver, data_interface, acm);

	usb_get_intf(control_interface);
	tty_register_device(acm_tty_driver, minor, &control_interface->dev);

	acm_table[minor] = acm;

	return 0;

alloc_fail5:
	usb_buffer_free(usb_dev, writesize * ACM_NWU, acm->write_buffer, acm->write_dma);
alloc_fail4:
	usb_buffer_free(usb_dev, readsize * ACM_NRU, acm->read_buffer, acm->read_dma);
alloc_fail3:
	usb_buffer_free(usb_dev, ctrlsize, acm->ctrl_buffer, acm->ctrl_dma);
alloc_fail2:
	kfree(acm);
alloc_fail:
	return -ENOMEM;
}

static void acm_disconnect(struct usb_interface *intf)
{
	struct acm *acm = usb_get_intfdata(intf);
	struct usb_device *usb_dev = interface_to_usbdev(intf);
	int i;

	if (!acm || !acm->dev) {
		dbg("disconnect on nonexisting interface");
		return;
	}

	mutex_lock(&open_mutex);
	if (!usb_get_intfdata(intf)) {
		mutex_unlock(&open_mutex);
		return;
	}
	if (acm->country_codes){
		device_remove_file(&acm->control->dev,
				&dev_attr_wCountryCodes);
		device_remove_file(&acm->control->dev,
				&dev_attr_iCountryCodeRelDate);
	}
	device_remove_file(&acm->control->dev, &dev_attr_bmCapabilities);
	acm->dev = NULL;
	usb_set_intfdata(acm->control, NULL);
	usb_set_intfdata(acm->data, NULL);

	tasklet_disable(&acm->urb_task);

	usb_kill_urb(&acm->ctrlurb);
	for (i = 0; i < ACM_NRU; i++)
		usb_kill_urb(&acm->read_urb[i]);
	for (i = 0; i < ACM_NWU; i++)
		usb_kill_urb(&acm->write_urb[i]);

	tasklet_disable(&acm->urb_task);

	flush_scheduled_work(); /* wait for acm_softint */

	usb_buffer_free(usb_dev, acm->ctrlsize, acm->ctrl_buffer, acm->ctrl_dma);
	usb_buffer_free(usb_dev, acm->writesize * ACM_NWU, acm->write_buffer, acm->write_dma);
	usb_buffer_free(usb_dev, acm->readsize * ACM_NRU, acm->read_buffer, acm->read_dma);
	usb_driver_release_interface(&acm_driver, intf == acm->control ? acm->data : intf);

	if (!acm->used) {
		acm_tty_unregister(acm);
		mutex_unlock(&open_mutex);
		return;
	}

	mutex_unlock(&open_mutex);

	if (acm->tty)
		tty_hangup(acm->tty);
}

/*
 * USB driver structure.
 */

static struct usb_device_id acm_ids[] = {
	/* quirky and broken devices */
	{ USB_DEVICE(0x0870, 0x0001), /* Metricom GS Modem */
	.driver_info = NO_UNION_NORMAL, /* has no union descriptor */
	},
	{ USB_DEVICE(0x0e8d, 0x0003), /* FIREFLY, MediaTek Inc; andrey.arapov@gmail.com */
	.driver_info = NO_UNION_NORMAL, /* has no union descriptor */
	},
	{ USB_DEVICE(0x0482, 0x0203), /* KYOCERA AH-K3001V */
	.driver_info = NO_UNION_NORMAL, /* has no union descriptor */
	},
	{ USB_DEVICE(0x079b, 0x000f), /* BT On-Air USB MODEM */
	.driver_info = NO_UNION_NORMAL, /* has no union descriptor */
	},
	{ USB_DEVICE(0x0ace, 0x1608), /* ZyDAS 56K USB MODEM */
	.driver_info = SINGLE_RX_URB, /* firmware bug */
	},
	{ USB_DEVICE(0x0ace, 0x1611), /* ZyDAS 56K USB MODEM - new version */
	.driver_info = SINGLE_RX_URB, /* firmware bug */
	},
	{ USB_DEVICE(0x22b8, 0x7000), /* Motorola Q Phone */
	.driver_info = NO_UNION_NORMAL, /* has no union descriptor */
	},

	/* control interfaces with various AT-command sets */
	{ USB_INTERFACE_INFO(USB_CLASS_COMM, USB_CDC_SUBCLASS_ACM,
		USB_CDC_ACM_PROTO_AT_V25TER) },
	{ USB_INTERFACE_INFO(USB_CLASS_COMM, USB_CDC_SUBCLASS_ACM,
		USB_CDC_ACM_PROTO_AT_PCCA101) },
	{ USB_INTERFACE_INFO(USB_CLASS_COMM, USB_CDC_SUBCLASS_ACM,
		USB_CDC_ACM_PROTO_AT_PCCA101_WAKE) },
	{ USB_INTERFACE_INFO(USB_CLASS_COMM, USB_CDC_SUBCLASS_ACM,
		USB_CDC_ACM_PROTO_AT_GSM) },
	{ USB_INTERFACE_INFO(USB_CLASS_COMM, USB_CDC_SUBCLASS_ACM,
		USB_CDC_ACM_PROTO_AT_3G	) },
	{ USB_INTERFACE_INFO(USB_CLASS_COMM, USB_CDC_SUBCLASS_ACM,
		USB_CDC_ACM_PROTO_AT_CDMA) },

	/* NOTE:  COMM/ACM/0xff is likely MSFT RNDIS ... NOT a modem!! */
	{ }
};

MODULE_DEVICE_TABLE (usb, acm_ids);

static struct usb_driver acm_driver = {
	.name =		"cdc_acm",
	.probe =	acm_probe,
	.disconnect =	acm_disconnect,
	.id_table =	acm_ids,
};

/*
 * TTY driver structures.
 */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static const struct tty_operations acm_ops = {
#else
static struct tty_operations acm_ops = {
#endif
	.open =			acm_tty_open,
	.close =		acm_tty_close,
	.write =		acm_tty_write,
	.write_room =		acm_tty_write_room,
	.ioctl =		acm_tty_ioctl,
	.throttle =		acm_tty_throttle,
	.unthrottle =		acm_tty_unthrottle,
	.chars_in_buffer =	acm_tty_chars_in_buffer,
	.break_ctl =		acm_tty_break_ctl,
	.set_termios =		acm_tty_set_termios,
	.tiocmget =		acm_tty_tiocmget,
	.tiocmset =		acm_tty_tiocmset,
};

/*
 * Init / exit.
 */

static int __init acm_init(void)
{
	int retval;
	acm_tty_driver = alloc_tty_driver(ACM_TTY_MINORS);
	if (!acm_tty_driver)
		return -ENOMEM;
	acm_tty_driver->owner = THIS_MODULE,
	acm_tty_driver->driver_name = "acm",
	acm_tty_driver->name = "ttyACM",
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	acm_tty_driver->devfs_name = "usb/acm/",
#endif
	acm_tty_driver->major = ACM_TTY_MAJOR,
	acm_tty_driver->minor_start = 0,
	acm_tty_driver->type = TTY_DRIVER_TYPE_SERIAL,
	acm_tty_driver->subtype = SERIAL_TYPE_NORMAL,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	acm_tty_driver->flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_DYNAMIC_DEV,
#else
	acm_tty_driver->flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS,
#endif
	acm_tty_driver->init_termios = tty_std_termios;
	acm_tty_driver->init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	tty_set_operations(acm_tty_driver, &acm_ops);

	retval = tty_register_driver(acm_tty_driver);
	if (retval) {
		put_tty_driver(acm_tty_driver);
		return retval;
	}

	retval = usb_register(&acm_driver);
	if (retval) {
		tty_unregister_driver(acm_tty_driver);
		put_tty_driver(acm_tty_driver);
		return retval;
	}

	info(DRIVER_VERSION ":" DRIVER_DESC);

	return 0;
}

static void __exit acm_exit(void)
{
	usb_deregister(&acm_driver);
	tty_unregister_driver(acm_tty_driver);
	put_tty_driver(acm_tty_driver);
}

module_init(acm_init);
module_exit(acm_exit);

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

