/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>


#if defined(_WIN32)
#define WIN32NATIVE
#endif
#if defined(WIN32NATIVE)
#include <windows.h>
#include <io.h>
#include <mmsystem.h>

static inline double dtime() {
	return (double)timeGetTime()/1.0e3;
}

void usleep(int t) {
	Sleep(1 + (t+999)/1000);
}


int set_serial(int fd) {
	HANDLE handle;
	DCB dcb;
	COMMTIMEOUTS cto;

	handle = (HANDLE)(void *)_get_osfhandle(fd);
	if (handle == INVALID_HANDLE_VALUE) {
		printf("get_osfhandle error\n");
		fflush(stdout);
		exit(1);
		//return -1;
	}
	if (!SetupComm(handle, 256, 256)) {
		printf("SetupComm error\n");
		fflush(stdout);
		return -1;
	}
	memset(&dcb, 0,  sizeof(DCB));
	dcb.DCBlength = sizeof(DCB);
	dcb.BaudRate = CBR_19200;
	dcb.fBinary = 1;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	if (!SetCommState(handle, &dcb)) {
		return -1;
	}
	memset(&cto, 0, sizeof(COMMTIMEOUTS));
	cto.ReadIntervalTimeout = 500;        // ms
	cto.ReadTotalTimeoutMultiplier = 500; // ms
	cto.ReadTotalTimeoutConstant = 500;   // ms
	SetCommTimeouts(handle, &cto);
	return 0;
}

#else
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>

static inline double dtime() {
	struct timeval tv;
	double t;

	gettimeofday(&tv,0);
	return (double)(unsigned long)tv.tv_sec + (double)tv.tv_usec/1.0e6;
}

int set_serial(int fd) {
	int speed;
	struct termios term;

	tcgetattr(fd, &term);
	//speed = B19200;
	speed = B3000000;
	cfsetispeed(&term,speed);
	cfsetospeed(&term,speed);
	cfmakeraw(&term);
	usleep(100000); /* 100 ms */
	tcsetattr(fd, TCSANOW, &term);
	/* setup done */
	return 0;
}

#ifndef O_BINARY
#define O_BINARY	0
#endif

#endif


#if defined(WIN32NATIVE)
char *port_name = "COM4";
#else
char *port_name = "/dev/ttyACM0";
#endif

dump_buf(char *s, unsigned char *buf, int len) {
	int i;
	fprintf(stderr,"%s:",s);
	for (i=0; i<len; i++) {
		fprintf(stderr,"%02x ",buf[i]);
	}
	fprintf(stderr,"\n");
}

usage() {
	fprintf(stderr,"perf_test --- test1 fw (echo) performance test\n");
	fprintf(stderr,"usage : perf_test [-P device_name] [-s io_size] [-l loop_count] [-o pipelinging_depth] \n");
}

main(int argc,char *argv[]) {
	int fd,c,i,r,size;
	int io_size,loop_cnt,ov_cnt;
	double st,et;
	int use_pipe = 0;
	unsigned char buf[1024],ibuf[1024],cbuf[1024];

	ov_cnt = 0;
	io_size = 4;
	loop_cnt = 1000;
	while ((c=getopt(argc,argv,"P:l:s:o:")) != EOF) {
	    switch(c) {
		case 'P':
			port_name = optarg;
			break;
		case 'l':
			loop_cnt = atoi(optarg);
			break;
		case 's':
			io_size = atoi(optarg);
			if (io_size >= 256) io_size = 256;
			break;
		case 'o':
			ov_cnt = atoi(optarg);
			break;
		default:
			usage();
			exit(1);
	    }
	}
	fd = open(port_name, O_RDWR|O_BINARY);
	if (fd < 0) {
		printf("can't open serial\n");
		usage();
		exit(1);
	}
	set_serial(fd);
	memset(buf, 0, sizeof(buf));

	printf("start --  size %3d overlap = %d loops %5d ",io_size,ov_cnt,loop_cnt);
	fflush(stdout);
	st = dtime();
	for (i=0; i< loop_cnt; i++) {
		memset(buf, (unsigned char)i, io_size);
		r = write(fd, buf, io_size);
		if (r != io_size) {
			fprintf(stderr, "error(1) w = %d io_size = %d\n",r,io_size);
			exit(1);
		}
		if (i >= ov_cnt) {
			size = io_size;
			while (size > 0) {
				r = read(fd, ibuf + io_size - size, size);
				if (r <= 0) {
					fprintf(stderr, "error(2) r = %d io_size = %d\n",r,io_size);
					exit(1);
				}
				size -= r;
			}
			memset(cbuf, (unsigned char)(i-ov_cnt), io_size);
			if (memcmp(cbuf, ibuf, io_size)) {
				fprintf(stderr, "error(3) compare error\n",r,io_size);
				dump_buf("w",cbuf,io_size);
				dump_buf("i",ibuf,io_size);
				exit(1);
			}
		}
	}
	for (i=0; i< ov_cnt; i++) {
		size = io_size;
		while (size > 0) {
			r = read(fd, ibuf + io_size - size, size);
			if (r <= 0) {
				fprintf(stderr, "error(4) r = %d io_size = %d\n",r,io_size);
				exit(1);
			}
			size -= r;
		}
		memset(cbuf, (unsigned char)(loop_cnt+i-ov_cnt), io_size);
		if (memcmp(cbuf, ibuf, io_size)) {
			fprintf(stderr, "error(5) compare error\n",r,io_size);
			dump_buf("w",cbuf,io_size);
			dump_buf("i",ibuf,io_size);
			exit(1);
		}
	}
	et = dtime();
	printf("end test  %5.3f sec\n", et - st);
}
