/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>


#if defined(_WIN32)
#define WIN32NATIVE
#endif
#if defined(WIN32NATIVE)
#include <windows.h>
#include <io.h>
#include <mmsystem.h>

static inline double dtime() {
	return (double)timeGetTime()/1.0e3;
}

void usleep(int t) {
	Sleep(1 + (t+999)/1000);
}


int set_serial(int fd) {
	HANDLE handle;
	DCB dcb;
	COMMTIMEOUTS cto;

	handle = (HANDLE)(void *)_get_osfhandle(fd);
	if (handle == INVALID_HANDLE_VALUE) {
		printf("get_osfhandle error\n");
		fflush(stdout);
		exit(1);
		//return -1;
	}
	if (!SetupComm(handle, 256, 256)) {
		printf("SetupComm error\n");
		fflush(stdout);
		return -1;
	}
	memset(&dcb, 0,  sizeof(DCB));
	dcb.DCBlength = sizeof(DCB);
	dcb.BaudRate = CBR_19200;
	dcb.fBinary = 1;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	if (!SetCommState(handle, &dcb)) {
		return -1;
	}
	memset(&cto, 0, sizeof(COMMTIMEOUTS));
	cto.ReadIntervalTimeout = 500;        // ms
	cto.ReadTotalTimeoutMultiplier = 500; // ms
	cto.ReadTotalTimeoutConstant = 500;   // ms
	SetCommTimeouts(handle, &cto);
	return 0;
}

#else
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>

static inline double dtime() {
	struct timeval tv;
	double t;

	gettimeofday(&tv,0);
	return (double)(unsigned long)tv.tv_sec + (double)tv.tv_usec/1.0e6;
}

int set_serial(int fd) {
	int speed;
	struct termios term;

	tcgetattr(fd, &term);
	speed = B19200;
	cfsetispeed(&term,speed);
	cfsetospeed(&term,speed);
	cfmakeraw(&term);
	usleep(100000); /* 100 ms */
	tcsetattr(fd, TCSANOW, &term);
	/* setup done */
	return 0;
}

#ifndef O_BINARY
#define O_BINARY	0
#endif

#endif


#if defined(WIN32NATIVE)
char *port_name = "COM4";
#else
char *port_name = "/dev/ttyACM0";
#endif

dump_buf(char *s, unsigned char *buf, int len) {
	int i;
	fprintf(stderr,"%s:",s);
	for (i=0; i<len; i++) {
		fprintf(stderr,"%02x ",buf[i]);
	}
	fprintf(stderr,"\n");
}

usage() {
	fprintf(stderr,"perf_test --- test2 fw (ADC/DAC simulation) performance test\n");
	fprintf(stderr,"usage : perf_test2 [-P device_name] [-s sampling Hz] [-l sample_count] [-r|-ro|-w|-x] \n");
	fprintf(stderr, "  -r: ADC (capture) simulation\n");
	fprintf(stderr, "     -ro: output time stamp option\n");
	fprintf(stderr, "  -w: DAC output simulation\n");
	fprintf(stderr, "  -x: ADC & DAC simulation\n");
}

main(int argc,char *argv[]) {
	int fd,c,i,r;
	double st,et;
	int rate;
	int rate_h,rate_l,rate_z;
	int loop_cnt,bytes,rlen;
	int out_data;
	int mode = 'r';
	int pre_bytes,rdone,ii;
	char cmd[1024];
	char ibuf[1024];

	rate = 0;
	loop_cnt = 0;
	out_data = 0;
	while ((c=getopt(argc,argv,"P:l:s:owrx")) != EOF) {
	    switch(c) {
		case 'P':
			port_name = optarg;
			break;
		case 'l': // sample_num
			loop_cnt = atoi(optarg);
			break;
		case 's': // sampling rate (HZ)
			rate = atoi(optarg);
			break;
		case 'o':
			out_data = 1;
			break;
		case 'w':
			mode = 'w';
			break;
		case 'r':
			mode = 'r';
			break;
		case 'x':
			mode = 'x';
			break;
		default:
			usage();
			exit(1);
	    }
	}
	if (loop_cnt <= 0) {
		loop_cnt = 1000;
	}
	if (rate <= 0) {
		rate = 1000; // 1kHz
	}
	rate_z = 1;
	rate_h = 1;
	rate_l = 8000000 / rate;
	if (rate_l >= 255) {
		rate_h ++;
		rate_l >>= 3;
		rate_z <<= 3;
	}
	if (rate_l >= 255) {
		rate_h ++;
		rate_l >>= 3;
		rate_z <<= 3;
	}
	if (rate_l >= 255) {
		rate_h ++;
		rate_l >>= 2;
		rate_z <<= 2;
	}
	if (rate_l >= 255) {
		rate_h ++;
		rate_l >>= 2;
		rate_z <<= 2;
	}
	rate = 8000000 / (rate_l * rate_z);
	if (loop_cnt >= 0xffff) {
		loop_cnt = 0xffff;
	}

	fd = open(port_name, O_RDWR|O_BINARY);
	if (fd < 0) {
		printf("can't open serial\n");
		usage();
		exit(1);
	}
	set_serial(fd);

	cmd[0] = 'S';
	write(fd, cmd, 1);
	r = read(fd, cmd, 7);
	if (r != 7) {
		printf("can't read signature  %d\n",r);
		exit(1);
	}
	cmd[7] = 0;
	if (memcmp(cmd, "USB162B", 7)) {
		printf("signature mismatch [%s]\n",cmd);
		exit(1);
	}
	cmd[0] = 'r';
	cmd[1] = rate_l;
	cmd[2] = rate_h;
	write(fd, cmd, 3);
	r = read(fd, cmd, 1);
	if ((r != 1) || (cmd[0] != '\r')) {
		printf("cmd r error\n");
		exit(1);
	}
	cmd[0] = 's';
	cmd[1] = loop_cnt;
	cmd[2] = loop_cnt>>8;
	write(fd, cmd, 3);
	r = read(fd, cmd, 1);
	if ((r != 1) || (cmd[0] != '\r')) {
		printf("cmd r error\n");
		exit(1);
	}
	printf("test %c start --  rate %6d loops %5d ",mode,rate,loop_cnt);
	fflush(stdout);
	if (mode == 'r') {
		cmd[0] = 'g';
		write(fd, cmd, 1);

		st = dtime();
		bytes = loop_cnt * 4;
		while (bytes > 0) {
			rlen = bytes;
			if (rlen > 1024) rlen = 1024;
			rdone = 0;
			while (rlen > 0) {
				r = read(fd, ibuf+rdone, rlen);
				if (r < 0) {
					printf("read error \n");
					exit(1);
				}
				rdone += r;
				rlen -= r;
			}
			if (out_data) for (ii=0; ii<rdone; ii+=4) {
				printf("\t%d\n", *(unsigned short *)(ibuf+ii));
			}
			bytes -= rdone;
		}
		et = dtime();
		printf("end test  %5.3f sec\n", et - st);
	} else if (mode == 'w') {
		cmd[0] = 'h'; // 'g' + 1
		write(fd, cmd, 1);

		st = dtime();
		bytes = loop_cnt * 4;
		while (bytes > 0) {
			rlen = bytes;
			if (rlen > 1024) rlen = 1024;
			rdone = 0;
			while (rlen > 0) {
				r = write(fd, ibuf+rdone, rlen);
				if (r < 0) {
					printf("write error \n");
					exit(1);
				}
				rdone += r;
				rlen -= r;
			}
			bytes -= rdone;
		}
		et = dtime();
		printf("end test  %5.3f sec\n", et - st);
	} else if (mode == 'x') {
#define XSIZE	1024
		cmd[0] = 'i'; // 'g' + 2
		write(fd, cmd, 1);

		st = dtime();
		bytes = loop_cnt * 4;
		pre_bytes = XSIZE;
		if (pre_bytes >= bytes) {
			pre_bytes = bytes;
		}
		r = write(fd, ibuf, pre_bytes);
		if (r <= 0) {
			printf("write error \n");
			exit(1);
		}
//fprintf(stderr, "x pre_bytes done\n");
//fflush(stderr);
		bytes -= pre_bytes;
		while (bytes > 0) {
			rlen = bytes;
			if (rlen > XSIZE) rlen = XSIZE;
			rdone = 0;
			while (rlen > 0) {
				r = write(fd, ibuf+rdone, rlen);
				if (r < 0) {
					printf("write error \n");
					exit(1);
				}
				rdone += r;
				rlen -= r;
			}
			rlen = rdone;
			rdone = 0;
			while (rlen > 0) {
				r = read(fd, ibuf+rdone, rlen);
				if (r < 0) {
					printf("read error \n");
					exit(1);
				}
				rdone += r;
				rlen -= r;
			}
			bytes -= rdone;
//fprintf(stderr, "x done %d bytes \n",loop_cnt * 4 - bytes);
//fflush(stderr);
		}
		rlen = pre_bytes;
		rdone = 0;
		while (rlen > 0) {
			r = read(fd, ibuf+rdone, rlen);
			if (r < 0) {
				printf("read error \n");
				exit(1);
			}
			rdone += r;
			rlen -= r;
		}
		et = dtime();
		printf("end test  %5.3f sec\n", et - st);
	}
}
